# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import pytest

from pylsp import uris
from pylsp.plugins.yapf_format import pylsp_format_document, pylsp_format_range
from pylsp.workspace import Document
from pylsp.text_edit import apply_text_edits

DOC_URI = uris.from_fs_path(__file__)
DOC = """A = [
    'h',   'w',

    'a'
      ]

B = ['h',


'w']
"""

GOOD_DOC = """A = ['hello', 'world']\n"""
FOUR_SPACE_DOC = """def hello():
    pass
"""


def test_format(workspace):
    doc = Document(DOC_URI, workspace, DOC)
    res = pylsp_format_document(workspace, doc, None)

    assert apply_text_edits(doc, res) == "A = ['h', 'w', 'a']\n\nB = ['h', 'w']\n"


def test_range_format(workspace):
    doc = Document(DOC_URI, workspace, DOC)

    def_range = {
        'start': {'line': 0, 'character': 0},
        'end': {'line': 4, 'character': 10}
    }
    res = pylsp_format_range(doc, def_range, None)

    # Make sure B is still badly formatted
    assert apply_text_edits(doc, res) == "A = ['h', 'w', 'a']\n\nB = ['h',\n\n\n'w']\n"


def test_no_change(workspace):
    doc = Document(DOC_URI, workspace, GOOD_DOC)
    assert not pylsp_format_document(workspace, doc, options=None)


def test_config_file(tmpdir, workspace):
    # a config file in the same directory as the source file will be used
    conf = tmpdir.join('.style.yapf')
    conf.write('[style]\ncolumn_limit = 14')
    src = tmpdir.join('test.py')
    doc = Document(uris.from_fs_path(src.strpath), workspace, DOC)

    res = pylsp_format_document(workspace, doc, options=None)

    # A was split on multiple lines because of column_limit from config file
    assert apply_text_edits(doc, res) == "A = [\n    'h', 'w',\n    'a'\n]\n\nB = ['h', 'w']\n"


@pytest.mark.parametrize('newline', ['\r\n'])
def test_line_endings(workspace, newline):
    doc = Document(DOC_URI, workspace, f'import os;import sys{2 * newline}dict(a=1)')
    res = pylsp_format_document(workspace, doc, options=None)

    assert apply_text_edits(doc, res) == f'import os{newline}import sys{2 * newline}dict(a=1){newline}'


def test_format_with_tab_size_option(workspace):
    doc = Document(DOC_URI, workspace, FOUR_SPACE_DOC)
    res = pylsp_format_document(workspace, doc, {"tabSize": "8"})

    assert apply_text_edits(doc, res) == FOUR_SPACE_DOC.replace("    ", "        ")


def test_format_with_insert_spaces_option(workspace):
    doc = Document(DOC_URI, workspace, FOUR_SPACE_DOC)
    res = pylsp_format_document(workspace, doc, {"insertSpaces": False})

    assert apply_text_edits(doc, res) == FOUR_SPACE_DOC.replace("    ", "\t")


def test_format_with_yapf_specific_option(workspace):
    doc = Document(DOC_URI, workspace, FOUR_SPACE_DOC)
    res = pylsp_format_document(workspace, doc, {"USE_TABS": True})

    assert apply_text_edits(doc, res) == FOUR_SPACE_DOC.replace("    ", "\t")


def test_format_returns_text_edit_per_line(workspace):
    single_space_indent = """def wow():
 log("x")
 log("hi")"""
    doc = Document(DOC_URI, workspace, single_space_indent)
    res = pylsp_format_document(workspace, doc, options=None)

    # two removes and two adds
    assert len(res) == 4
    assert res[0]['newText'] == ""
    assert res[1]['newText'] == ""
    assert res[2]['newText'] == "    log(\"x\")\n"
    assert res[3]['newText'] == "    log(\"hi\")\n"



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import pytest

from pylsp import uris
from pylsp.plugins.autopep8_format import pylsp_format_document, pylsp_format_range
from pylsp.workspace import Document

DOC_URI = uris.from_fs_path(__file__)
DOC = """a =    123




def func():
    pass
"""

GOOD_DOC = """A = ['hello', 'world']\n"""

INDENTED_DOC = """def foo():
    print('asdf',
    file=None
    )

bar = { 'foo': foo
}
"""

CORRECT_INDENTED_DOC = """def foo():
    print('asdf',
          file=None
          )


bar = {'foo': foo
       }
"""


def test_format(config, workspace):
    doc = Document(DOC_URI, workspace, DOC)
    res = pylsp_format_document(config, workspace, doc, options=None)

    assert len(res) == 1
    assert res[0]['newText'] == "a = 123\n\n\ndef func():\n    pass\n"


def test_range_format(config, workspace):
    doc = Document(DOC_URI, workspace, DOC)

    def_range = {
        'start': {'line': 0, 'character': 0},
        'end': {'line': 2, 'character': 0}
    }
    res = pylsp_format_range(config, workspace, doc, def_range, options=None)

    assert len(res) == 1

    # Make sure the func is still badly formatted
    assert res[0]['newText'] == "a = 123\n\n\n\n\ndef func():\n    pass\n"


def test_no_change(config, workspace):
    doc = Document(DOC_URI, workspace, GOOD_DOC)
    assert not pylsp_format_document(config, workspace, doc, options=None)


def test_hanging_indentation(config, workspace):
    doc = Document(DOC_URI, workspace, INDENTED_DOC)
    res = pylsp_format_document(config, workspace, doc, options=None)

    assert len(res) == 1
    assert res[0]['newText'] == CORRECT_INDENTED_DOC


@pytest.mark.parametrize('newline', ['\r\n', '\r'])
def test_line_endings(config, workspace, newline):
    doc = Document(DOC_URI, workspace, f'import os;import sys{2 * newline}dict(a=1)')
    res = pylsp_format_document(config, workspace, doc, options=None)

    assert res[0]['newText'] == f'import os{newline}import sys{2 * newline}dict(a=1){newline}'



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import pytest
from pylsp import IS_WIN


unix_only = pytest.mark.skipif(IS_WIN, reason="Unix only")
windows_only = pytest.mark.skipif(not IS_WIN, reason="Windows only")



# Copyright 2022- Python Language Server Contributors.

from typing import Dict, List
from unittest.mock import Mock

import jedi
import parso
import pytest

from pylsp import lsp, uris
from pylsp.config.config import Config
from pylsp.plugins.rope_autoimport import _get_score, _should_insert, get_names
from pylsp.plugins.rope_autoimport import \
    pylsp_completions as pylsp_autoimport_completions
from pylsp.plugins.rope_autoimport import pylsp_initialize
from pylsp.workspace import Workspace

DOC_URI = uris.from_fs_path(__file__)


@pytest.fixture(scope="session")
def autoimport_workspace(tmp_path_factory) -> Workspace:
    "Special autoimport workspace. Persists across sessions to make in-memory sqlite3 database fast."
    workspace = Workspace(uris.from_fs_path(str(tmp_path_factory.mktemp("pylsp"))), Mock())
    workspace._config = Config(workspace.root_uri, {}, 0, {})
    workspace._config.update({"rope_autoimport": {"memory": True, "enabled": True}})
    pylsp_initialize(workspace._config, workspace)
    yield workspace
    workspace.close()


# pylint: disable=redefined-outer-name
@pytest.fixture
def completions(config: Config, autoimport_workspace: Workspace, request):
    document, position = request.param
    com_position = {"line": 0, "character": position}
    autoimport_workspace.put_document(DOC_URI, source=document)
    doc = autoimport_workspace.get_document(DOC_URI)
    yield pylsp_autoimport_completions(config, autoimport_workspace, doc, com_position)
    autoimport_workspace.rm_document(DOC_URI)


def should_insert(phrase: str, position: int):
    expr = parso.parse(phrase)
    word_node = expr.get_leaf_for_position((1, position))
    return _should_insert(expr, word_node)


def check_dict(query: Dict, results: List[Dict]) -> bool:
    for result in results:
        if all(result[key] == query[key] for key in query.keys()):
            return True
    return False


@pytest.mark.parametrize("completions", [("""pathli """, 6)], indirect=True)
def test_autoimport_completion(completions):
    assert completions
    assert check_dict(
        {
            "label": "pathlib",
            "kind": lsp.CompletionItemKind.Module
        }, completions)


@pytest.mark.parametrize("completions", [("""import """, 7)], indirect=True)
def test_autoimport_import(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""pathlib""", 2)], indirect=True)
def test_autoimport_pathlib(completions):
    assert completions[0]["label"] == "pathlib"

    start = {"line": 0, "character": 0}
    edit_range = {"start": start, "end": start}
    assert completions[0]["additionalTextEdits"] == [{
        "range":
        edit_range,
        "newText":
        "import pathlib\n"
    }]


@pytest.mark.parametrize("completions", [("""import test\n""", 10)],
                         indirect=True)
def test_autoimport_import_with_name(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""def func(s""", 10)],
                         indirect=True)
def test_autoimport_function(completions):

    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""class Test""", 10)],
                         indirect=True)
def test_autoimport_class(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""\n""", 0)], indirect=True)
def test_autoimport_empty_line(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""class Test(NamedTupl):""", 20)],
                         indirect=True)
def test_autoimport_class_complete(completions):
    assert len(completions) > 0


@pytest.mark.parametrize("completions", [("""class Test(NamedTupl""", 20)],
                         indirect=True)
def test_autoimport_class_incomplete(completions):
    assert len(completions) > 0


@pytest.mark.parametrize("completions", [("""def func(s:Lis""", 12)],
                         indirect=True)
def test_autoimport_function_typing(completions):
    assert len(completions) > 0
    assert check_dict({"label": "List"}, completions)


@pytest.mark.parametrize("completions", [("""def func(s : Lis ):""", 16)],
                         indirect=True)
def test_autoimport_function_typing_complete(completions):
    assert len(completions) > 0
    assert check_dict({"label": "List"}, completions)


@pytest.mark.parametrize("completions",
                         [("""def func(s : Lis ) -> Generat:""", 29)],
                         indirect=True)
def test_autoimport_function_typing_return(completions):
    assert len(completions) > 0
    assert check_dict({"label": "Generator"}, completions)


def test_autoimport_defined_name(config, workspace):
    document = """List = "hi"\nLis"""
    com_position = {"line": 1, "character": 3}
    workspace.put_document(DOC_URI, source=document)
    doc = workspace.get_document(DOC_URI)
    completions = pylsp_autoimport_completions(config, workspace, doc,
                                               com_position)
    workspace.rm_document(DOC_URI)
    assert not check_dict({"label": "List"}, completions)


# This test has several large issues.
# 1. autoimport relies on its sources being written to disk. This makes testing harder
# 2. the hook doesn't handle removed files
# 3. The testing framework cannot access the actual autoimport object so it cannot clear the cache
# def test_autoimport_update_module(config: Config, workspace: Workspace):
#     document2 = "SomethingYouShouldntWrite = 1"
#     document = """SomethingYouShouldntWrit"""
#     com_position = {
#         "line": 0,
#         "character": 3,
#     }
#     doc2_path = workspace.root_path + "/test_file_no_one_should_write_to.py"
#     if os.path.exists(doc2_path):
#         os.remove(doc2_path)
#     DOC2_URI = uris.from_fs_path(doc2_path)
#     workspace.put_document(DOC_URI, source=document)
#     doc = workspace.get_document(DOC_URI)
#     completions = pylsp_autoimport_completions(config, workspace, doc, com_position)
#     assert len(completions) == 0
#     with open(doc2_path, "w") as f:
#         f.write(document2)
#     workspace.put_document(DOC2_URI, source=document2)
#     doc2 = workspace.get_document(DOC2_URI)
#     pylsp_document_did_save(config, workspace, doc2)
#     assert check_dict({"label": "SomethingYouShouldntWrite"}, completions)
#     workspace.put_document(DOC2_URI, source="\n")
#     doc2 = workspace.get_document(DOC2_URI)
#     os.remove(doc2_path)
#     pylsp_document_did_save(config, workspace, doc2)
#     completions = pylsp_autoimport_completions(config, workspace, doc, com_position)
#     assert len(completions) == 0
#     workspace.rm_document(DOC_URI)


class TestShouldInsert:

    def test_dot(self):

        assert not should_insert("""str.""", 4)

    def test_dot_partial(self):

        assert not should_insert("""str.metho\n""", 9)

    def test_comment(self):
        assert not should_insert("""#""", 1)

    def test_comment_indent(self):

        assert not should_insert("""    # """, 5)

    def test_from(self):
        assert not should_insert("""from """, 5)
        assert should_insert("""from """, 4)


def test_sort_sources():
    result1 = _get_score(1, "import pathlib", "pathlib", "pathli")
    result2 = _get_score(2, "import pathlib", "pathlib", "pathli")
    assert result1 < result2


def test_sort_statements():
    result1 = _get_score(2, "from importlib_metadata import pathlib",
                         "pathlib", "pathli")
    result2 = _get_score(2, "import pathlib", "pathlib", "pathli")
    assert result1 > result2


def test_sort_both():
    result1 = _get_score(3, "from importlib_metadata import pathlib",
                         "pathlib", "pathli")
    result2 = _get_score(2, "import pathlib", "pathlib", "pathli")
    assert result1 > result2


def test_get_names():
    source = """
    from a import s as e
    import blah, bleh
    hello = "str"
    a, b = 1, 2
    def someone():
        soemthing
    class sfa:
        sfiosifo
    """
    results = get_names(jedi.Script(code=source))
    assert results == set(
        ["blah", "bleh", "e", "hello", "someone", "sfa", "a", "b"])



import json
import sys
from argparse import ArgumentParser, FileType


def describe_array(prop: dict) -> str:
    extra = ""
    if "items" in prop:
        unique_qualifier = ""
        if "uniqueItems" in prop:
            unique_qualifier = "unique" if prop["uniqueItems"] else "non-unique"
        item_type = describe_type(prop["items"])
        extra = " ".join(filter(bool, ["of", unique_qualifier, item_type, "items"]))
    return extra


def describe_number(prop: dict) -> str:
    extra = []
    if "minimum" in prop:
        extra.append(f">= {prop['minimum']}")
    if "maximum" in prop:
        extra.append(f"<= {prop['maximum']}")
    return ",".join(extra)


EXTRA_DESCRIPTORS = {
    "array": describe_array,
    "number": describe_number,
}


def describe_type(prop: dict) -> str:
    prop_type = prop["type"]
    types = prop_type if isinstance(prop_type, list) else [prop_type]
    if "null" in types:
        types.remove("null")
    if len(types) == 1:
        prop_type = types[0]
    parts = [f"`{prop_type}`"]
    for option in types:
        if  option in EXTRA_DESCRIPTORS:
            parts.append(EXTRA_DESCRIPTORS[option](prop))
    if "enum" in prop:
        allowed_values = [f"`{value!r}`" for value in prop["enum"]]
        parts.append("(one of: " + ", ".join(allowed_values) + ")")
    return " ".join(parts)


def convert_schema(schema: dict, source: str = None) -> str:
    lines = [
        f"# {schema['title']}",
        schema["description"],
        "",
        "| **Configuration Key** | **Type** | **Description** | **Default** ",
        "|----|----|----|----|",
    ]
    for key, prop in schema["properties"].items():
        description = prop.get("description", "")
        default = json.dumps(prop.get("default", ""))
        lines.append(
            f"| `{key}` | {describe_type(prop)} | {description} | `{default}` |"
        )

    if source:
        lines.append(
            f"\nThis documentation was generated from `{source}`."
            " Please do not edit this file directly."
        )

    # ensure empty line at the end
    lines.append("")

    return "\n".join(lines)


def main(argv):
    parser = ArgumentParser()
    parser.add_argument("schema", type=FileType())
    parser.add_argument("markdown", type=FileType("w+"), default=sys.stdout)
    arguments = parser.parse_args(argv[1:])
    schema = json.loads(arguments.schema.read())
    markdown = convert_schema(schema, source=arguments.schema.name)
    arguments.markdown.write(markdown)


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3

# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

from setuptools import setup, find_packages


if __name__ == "__main__":
    setup(
        name="python-lsp-server",  # to allow GitHub dependency tracking work
    )



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

from pylsp.text_edit import OverLappingTextEditException, apply_text_edits
from pylsp import uris

DOC_URI = uris.from_fs_path(__file__)


def test_apply_text_edits_insert(pylsp):
    pylsp.workspace.put_document(DOC_URI, '012345678901234567890123456789')
    test_doc = pylsp.workspace.get_document(DOC_URI)

    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 0
            },
            "end": {
                "line": 0,
                "character": 0
            }
        },
        "newText": "Hello"
    }]) == 'Hello012345678901234567890123456789'
    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 1
            },
            "end": {
                "line": 0,
                "character": 1
            }
        },
        "newText": "Hello"
    }]) == '0Hello12345678901234567890123456789'
    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 1
            },
            "end": {
                "line": 0,
                "character": 1
            }
        },
        "newText": "Hello"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 1
            },
            "end": {
                "line": 0,
                "character": 1
            }
        },
        "newText": "World"
    }]) == '0HelloWorld12345678901234567890123456789'
    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 2
            },
            "end": {
                "line": 0,
                "character": 2
            }
        },
        "newText": "One"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 1
            },
            "end": {
                "line": 0,
                "character": 1
            }
        },
        "newText": "Hello"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 1
            },
            "end": {
                "line": 0,
                "character": 1
            }
        },
        "newText": "World"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 2
            },
            "end": {
                "line": 0,
                "character": 2
            }
        },
        "newText": "Two"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 2
            },
            "end": {
                "line": 0,
                "character": 2
            }
        },
        "newText": "Three"
    }]) == '0HelloWorld1OneTwoThree2345678901234567890123456789'


def test_apply_text_edits_replace(pylsp):
    pylsp.workspace.put_document(DOC_URI, '012345678901234567890123456789')
    test_doc = pylsp.workspace.get_document(DOC_URI)

    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 3
            },
            "end": {
                "line": 0,
                "character": 6
            }
        },
        "newText": "Hello"
    }]) == '012Hello678901234567890123456789'
    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 3
            },
            "end": {
                "line": 0,
                "character": 6
            }
        },
        "newText": "Hello"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 6
            },
            "end": {
                "line": 0,
                "character": 9
            }
        },
        "newText": "World"
    }]) == '012HelloWorld901234567890123456789'
    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 3
            },
            "end": {
                "line": 0,
                "character": 6
            }
        },
        "newText": "Hello"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 6
            },
            "end": {
                "line": 0,
                "character": 6
            }
        },
        "newText": "World"
    }]) == '012HelloWorld678901234567890123456789'
    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 6
            },
            "end": {
                "line": 0,
                "character": 6
            }
        },
        "newText": "World"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 3
            },
            "end": {
                "line": 0,
                "character": 6
            }
        },
        "newText": "Hello"
    }]) == '012HelloWorld678901234567890123456789'
    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 0,
                "character": 3
            },
            "end": {
                "line": 0,
                "character": 3
            }
        },
        "newText": "World"
    }, {
        "range": {
            "start": {
                "line": 0,
                "character": 3
            },
            "end": {
                "line": 0,
                "character": 6
            }
        },
        "newText": "Hello"
    }]) == '012WorldHello678901234567890123456789'


def test_apply_text_edits_overlap(pylsp):
    pylsp.workspace.put_document(DOC_URI, '012345678901234567890123456789')
    test_doc = pylsp.workspace.get_document(DOC_URI)

    did_throw = False
    try:
        apply_text_edits(test_doc, [{
            "range": {
                "start": {
                    "line": 0,
                    "character": 3
                },
                "end": {
                    "line": 0,
                    "character": 6
                }
            },
            "newText": "Hello"
        }, {
            "range": {
                "start": {
                    "line": 0,
                    "character": 3
                },
                "end": {
                    "line": 0,
                    "character": 3
                }
            },
            "newText": "World"
        }])
    except OverLappingTextEditException:
        did_throw = True

    assert did_throw

    did_throw = False

    try:
        apply_text_edits(test_doc, [{
            "range": {
                "start": {
                    "line": 0,
                    "character": 3
                },
                "end": {
                    "line": 0,
                    "character": 6
                }
            },
            "newText": "Hello"
        }, {
            "range": {
                "start": {
                    "line": 0,
                    "character": 4
                },
                "end": {
                    "line": 0,
                    "character": 4
                }
            },
            "newText": "World"
        }])
    except OverLappingTextEditException:
        did_throw = True

    assert did_throw


def test_apply_text_edits_multiline(pylsp):
    pylsp.workspace.put_document(DOC_URI, '0\n1\n2\n3\n4')
    test_doc = pylsp.workspace.get_document(DOC_URI)

    assert apply_text_edits(test_doc, [{
        "range": {
            "start": {
                "line": 2,
                "character": 0
            },
            "end": {
                "line": 3,
                "character": 0
            }
        },
        "newText": "Hello"
    }, {
        "range": {
            "start": {
                "line": 1,
                "character": 1
            },
            "end": {
                "line": 1,
                "character": 1
            }
        },
        "newText": "World"
    }]) == '0\n1World\nHello3\n4'
